package graph;

import java.util.*;
import java.awt.Color;

// Represents a graph (a collection of points) that can be displayed to a user
public abstract class Graph {
    // The resolution of the resulting image
    public static final int RESOLUTION = 1500;
    // What the baseline for the axes bounds should be
    public static final int AXIS_BOUND = 0;
    // How many pixels to add to the edges of the image
    // (such that points at the very end aren't cut off)
    public static final int BUFFER = 10;

    // Adds a graphable element to this graph
    public abstract void add(Graphable g);

    // Returns the color of the provided graphable element, null if it's not present
    public abstract Color getColor(Graphable g);

    // Returns all graphable elements within this graph
    public abstract Set<Graphable> getElements();

    // Returns all the points for this graph
    public List<Point> getPoints() {
        List<Point> ret = new ArrayList<>();
        for (Graphable g : this.getElements()) {
            if (!(g instanceof Point) && !(g instanceof Line)) {
                throw new IllegalStateException("Unexpected Graphable object: " + g);
            }
            
            if (g instanceof Point) {
                ret.add((Point)g);
            } else if (g instanceof Line) {
                ret.addAll(((Line)g).getPoints());
            }
        }
        return ret;
    }

    // Behavior: Saves this graph as an image to a file named 'fileName'
    // Exceptions:
    //      IllegalStateException if the RESOLUTION is too small for the graph to be drawn
    // Parameters:
    //      fileName - The name of the file the resulting graph should save to
    public void save(String fileName) {
        Picture pic = new Picture(RESOLUTION, RESOLUTION);
        pic.setOriginLowerLeft();
        Color[][] pixels = pic.getPixels();

        // Find the corners of the graphable region
        Point upperRight = new Point(AXIS_BOUND, AXIS_BOUND);
        Point lowerLeft = new Point(-AXIS_BOUND, -AXIS_BOUND);
        for (Point p : getPoints()) {
            upperRight.setLocation((long)Math.ceil(Math.max(upperRight.getX(), p.getX())),
                                   (long)Math.ceil(Math.max(upperRight.getY(), p.getY())));
            lowerLeft.setLocation((long)Math.floor(Math.min(lowerLeft.getX(), p.getX())),
                                  (long)Math.floor(Math.min(lowerLeft.getY(), p.getY())));
        }
            
        // Ratios need to be > 2 otherwise background will be entirely gray
        if (getRatio(lowerLeft, upperRight, true) <= 2 ||
            getRatio(lowerLeft, upperRight, false) <= 2) {
            
            throw new IllegalStateException("Points cover span of area too large - " +
                                            "increase RESOLUTION class constant");
        }

        // Draw all the grid lines
        drawGrids(pixels, lowerLeft, upperRight);

        // Draw the elements
        for (Graphable g : this.getElements()) {
            if (!(g instanceof Point) && !(g instanceof Line)) {
                throw new IllegalStateException("Unexpected Graphable object: " + g);
            }

            Color c = this.getColor(g);
            if (g instanceof Point) {
                drawSquare(pixels, unitToPixels(lowerLeft, upperRight, (Point)g), 5, c);
            } else if (g instanceof Line) {
                List<Point> linePoints = ((Line)g).getPoints();
                if (!linePoints.isEmpty()) {
                    drawSquare(pixels,
                               unitToPixels(lowerLeft, upperRight, linePoints.get(0)),
                               5, c);
                    for (int i = 1; i < linePoints.size(); i++) {
                        drawSquare(pixels,
                                   unitToPixels(lowerLeft, upperRight, linePoints.get(i)),
                                   5, c);
                        drawLine(pixels,
                                unitToPixels(lowerLeft, upperRight, linePoints.get(i - 1)),
                                unitToPixels(lowerLeft, upperRight, linePoints.get(i)),
                                2, c);
                    }
                }
            }
        }
        
        // Save the resulting image
        pic.setPixels(pixels);
        pic.save(fileName);
    }

    // Behavior: Draws the gray grid lines and the black axis lines within the provided 2d
    //           array of color 'pixels'.
    // Parameters:
    //      pixels - The 2d array to fill to draw grid lines on
    //      ll - The lower left of the region being graphed
    //      ur - The upper right of the region being graphed
    private void drawGrids(Color[][] pixels, Point ll, Point ur) {
        // Fill the background to be white;
        fill(pixels, Color.WHITE);
        
        // Draw vertical grid lines
        for (int i = (int)ll.getX(); i <= ur.getX(); i++) {
            drawVerticalLine(pixels, unitToPixels(getRatio(ll, ur, true), ll.getX(), i),
                             1, Color.GRAY);
        }

        // Draw horizontal grid lines
        for (int i = (int)ll.getY(); i <= ur.getY(); i++) {
            drawHorizontalLine(pixels, unitToPixels(getRatio(ll, ur, false), ll.getY(), i),
                               1, Color.GRAY);
        }
        
        // Draw the axis lines
        Point pixelOrigin = unitToPixels(ll, ur, new Point());
        drawVerticalLine(pixels, pixelOrigin.getX(), 3, Color.BLACK);
        drawHorizontalLine(pixels, pixelOrigin.getY(), 3, Color.BLACK);
    }

    // Behavior: Determines the appropriate axis ratio depending on the size of the region
    //           being graphed
    // Parameters:
    //      ll - The lower left of the region being graphed
    //      ur - The upper right of the region being graphed
    //      horizontal - Whether or not to return the horizontal ratio
    // Returns: The appropriate axis ratio for the region being graphed
    private double getRatio(Point ll, Point ur, boolean horizontal) {
        if (horizontal) {
            return (RESOLUTION - (2 * BUFFER)) / (double)(ur.getX() - ll.getX());
        } else {
            return (RESOLUTION - (2 * BUFFER)) / (double)(ur.getY() - ll.getY());
        }
    }

    // Behavior: Converts a value 'u' from unit space to pixel-space
    // Parameters:
    //      ratio - The ratio for whatever axis we're converting on
    //      start - The starting (minimum) value for the axis we're converting on
    //      u - The actual unit value to convert to pixel-space
    // Returns: The appropriate int pixel-space coordinate for 'u'
    private int unitToPixels(double ratio, double start, double u) {
        return (int)(BUFFER + Math.round((u - start) * ratio));
    }

    // Behavior: Converts a Point 'p' from unit space to pixel-space
    // Parameters:
    //      ll - The starting (lower left) value for the axes we're converting to
    //      ur - The ending (upper right) value for the axes we're converting to
    //      p - The actual point to convert to pixel-space
    // Returns: The appropriate Point pixel-space coordinate for 'p'
    private Point unitToPixels(Point ll, Point ur, Point p) {
        return new Point(unitToPixels(getRatio(ll, ur, true), ll.getX(), p.getX()),
                         unitToPixels(getRatio(ll, ur, false), ll.getY(), p.getY()));
    }

    // Behavior: Fills the provided 2d array of color 'pixels' with the provided Color 'c'
    // Parameters:
    //      pixels - The 2d array to fill
    //      c - The color to fill with
    private void fill(Color[][] pixels, Color c) {
        for (int i = 0; i < pixels.length; i++) {
            for (int j = 0; j < pixels[i].length; j++) {
                pixels[i][j] = c;
            }
        }
    }

    // Behavior: Draws a square within the provided 2d array of Color 'pixels' centered at Point
    //           'center'. The square will have a radius 'radius' and will be colored Color 'c'
    // Exceptions:
    //      IllegalArgumentException if the provided point is outside the bounds of the 2d array
    // Parameters:
    //      pixels - The 2d array to draw the square within
    //      center - The center of the square to draw
    //      radius - The radius (distance from center to edge) of the square to draw
    //      c - The color to make the square
    private void drawSquare(Color[][] pixels, Point center, int radius, Color c) {
        if (center.getX() < 0 || center.getX() >= pixels.length ||
            center.getY() < 0 || center.getY() >= pixels[0].length) {
                throw new IllegalArgumentException("Point " + center + " out of bounds for " +
                                                   "axes [0, " + RESOLUTION + ")");
        }
        
        for (int i = (int)Math.max(0, center.getY() - radius);
                 i <= (int)Math.min(pixels.length - 1, center.getY() + radius);
                 i++) {
            for (int j = (int)Math.max(0, center.getX() - radius);
                     j <= (int)Math.min(pixels.length - 1, center.getX() + radius);
                     j++) {
                pixels[i][j] = c;
            }
        }
    }

    // Behavior: Draws a line within the provided 2d array or Color 'pixels' between the provided
    //           'start' and 'end' points with width 'width' and Color 'c'. Warning, the more vertical
    //           the line, the less line-like the result will be.
    // Parameters:
    //      pixels - The 2d array to draw the line within
    //      start - The starting point of the line (in pixel space)
    //      end - The ending point of the line (in pixels space)
    //      width - The width of the line in pixels
    //      c - The color to make the line
    private void drawLine(Color[][] pixels, Point start, Point end, int width, Color c) {
        double slope = (double)(end.getY() - start.getY()) / (end.getX() - start.getX());
        int bound = (int)(Math.floor(width / 2) + Math.abs(slope));
        
        for (int x = (int)start.getX(); x <= end.getX(); x++) {
            int y = (int)(Math.round((x - start.getX()) * slope) + start.getY());
            for (int i = (int)Math.max(0, y - bound); i <= Math.min(pixels.length - 1, y + bound); i++) {
                pixels[i][x] = c;
            }
        }
    }

    // Behavior: Draws a horizontal line within the provided 2d array or Color 'pixels' at
    //           value 'y' with width 'width' and Color 'c'
    // Parameters:
    //      pixels - The 2d array to draw the line within
    //      y - The pixel-space y value to draw the line at
    //      width - The width of the line in pixels
    //      c - The color to make the line
    private void drawHorizontalLine(Color[][] pixels, double y, int width, Color c) {
        int bound = (int)(Math.floor(width / 2));
        for (int i = (int)Math.max(0, y - bound); i <= Math.min(pixels.length - 1, y + bound); i++) {
            for (int j = 0; j < pixels[0].length; j++) {
                pixels[i][j] = c;
            }
        }
    }

    // Behavior: Draws a vertical line within the provided 2d array or Color 'pixels' at
    //           value 'x' with width 'width' and Color 'c'
    // Parameters:
    //      pixels - The 2d array to draw the line within
    //      x - The pixel-space x value to draw the line at
    //      width - The width of the line in pixels
    //      c - The color to make the line
    private void drawVerticalLine(Color[][] pixels, double x, int width, Color c) {
        int bound = (int)(Math.floor(width / 2));
        for (int i = 0; i < pixels.length; i++) {
            for (int j = (int)Math.max(0, x - bound); j <= Math.min(pixels[0].length - 1, x + bound); j++) {
                pixels[i][j] = c;
            }
        }
    }
}
